/* generated by Makefile */
extern const struct hw_descriptor dv_cfi_descriptor[];
extern const struct hw_descriptor dv_core_descriptor[];
extern const struct hw_descriptor dv_pal_descriptor[];
extern const struct hw_descriptor dv_glue_descriptor[];
const struct hw_descriptor * const hw_descriptors[] = {
  dv_cfi_descriptor,
  dv_core_descriptor,
  dv_pal_descriptor,
  dv_glue_descriptor,
  NULL,
};
